
#--- import libraries
import scipy
from scipy import ndimage

from pylab import *
from multiprocessing import Process,Pool

import numpy as np
import nosetrackerv1 as nt
import nosetracker_config as ntc

import os
import shutil
import time
import math
import random
import glob

import skimage
from skimage import img_as_float
from skimage.morphology import binary_closing, disk
from skimage.filters.rank import entropy #bug? if above certain packages, won't properly import, keep low

###

def multicorewurm(startf,endf,blockc,pcrop,directory,targetsize,bestedge,Globalstart):
	#---generate list of parallel processable chunks
	endf = endf + 1 #to generate the correct blocks (last frame specifically)
	chunks = [startf]
	
	#generate new blocks, if the range is bigger than a single block
	try:
		#find the end of the first blockend after start if possible
		for i in range(0,len(blockc)):
			if ((startf < blockc[i]) and (blockc[i] < endf)):
				chunks.append(blockc[i])
				index = i
				break
		
		#genereate successive blocks from blocklist
		for i in range(index,len(blockc)-1):
			if (blockc[i+1] < endf):
				chunks.append(blockc[i])
				chunks.append(blockc[i+1])
			else:
				break
			
		#generate last chunk from blocklist till end
		chunks.append(chunks[-1])
		chunks.append(endf)

	except:
		chunks.append(endf)
	
	total_chunks = int(len(chunks)/2)
	
	#spawn as much processes as blocks are available or until endframe is reached
	for i in range(0,int(len(chunks)/2)):
		blockstart = chunks[i*2]
		blockend = chunks[i*2+1]
		
		chunknumber = i
		
		pool.apply_async(wurmcal, args=(blockstart,blockend,pcrop,directory,targetsize,bestedge,chunknumber,Globalstart,total_chunks))
	
	return(total_chunks)

def getthreshold (nparray,targetsize):
	#--- apply gauss filter
	wurmf = ndimage.gaussian_filter(nparray,1.5) #gaussfilterradius, 1.5 seems to be ok
	
	#set the thresholds for contrast to brightest/darkest pixel
	startthresh = wurmf.min()
	endthresh = wurmf.max()
	
	#initialize stuff
	size = 0
	supertired = 0
	breakentropy = 0 #check to avoid second comparison if a zero was encountered in the first
	c = (random.random() * (endthresh-startthresh)) + startthresh #still better than zero
	bestobj = np.zeros((3))
	
	###### monte carlo approach #####
	while (not(targetsize * ntc.ct_upper_size_limit < size < targetsize * ntc.ct_upper_size_limit)): #size should be within 0.5% of target size
		wurmfertig = wurmf < c
		#--- label all objects
		label_wurm, num = ndimage.label(wurmfertig) #label alle objekte die nach dem threshold noch übrig sind
	
		#------ calculate areas, delete all but biggest --> könnte noch optimiert werden
		area = ndimage.sum(wurmfertig, label_wurm, index=arange(label_wurm.max() + 1)) #sum addiert alle zu einem label gehörenden positionen (binäre 1en) zusammen. wurmfertig ist binäres array, label_wurm enthält die label der areas in wurmfertig. index bestimmt über welche bereiche des arrays die summen gebildet werden sollen. in diesem fall die summe des binären arrays wurmfertig mit den Bereichen die gelabled wurden. und das von allen labeln (also label.max+1)
		
		#------ keep unaltered area-table --> optimize to do this only on last run(take latest threshold etc.)
		allareas = area.copy()
		
		#------ take biggest chunk
		largest = argmax(area)
		for i in range(0,label_wurm.max() + 1): # in dem 1D array area werden nun alle objekte die nicht das größte sind gelich null gesetzt, so das nur das größte übrig bleibt
			if i != argmax(area):
				area[i] = 0

		size = max(area)
		
		#------ reduce range wherein random numbers are looked for
		if size > targetsize:
			endthresh = c
		if size < targetsize:
			startthresh = c

		c = (random.random() * (endthresh-startthresh)) + startthresh		
		
		#--- prevent algorithm from getting stuck if threshold too narrow
		supertired += 1

		if supertired >= ntc.max_attempts:
			break
			
	#------ determine chunk with lowest entropy (only works if at least 3 chunks)
	entro = entropy(nparray, disk(3))
	
	if (allareas.shape[0] >= 3): #check against number of objects
		for i in range(1,4): #check the 3 largest objects (not 0-3 because backwards)
			
			temp_allareas = allareas.copy() #make a copy of the area-array so we can modify it for entropy calc
			n_largest = temp_allareas.argsort()[-i] #gets the index of the n-largest([-i]-largest) object
			
			for j in range(0,label_wurm.max() + 1): #remove all but chosen object
				if j != n_largest:
					temp_allareas[j] = 0
			
			temp_size = max(temp_allareas)
			temp_allareas[argmax(temp_allareas)] = 1
			temp_areaImg = temp_allareas[label_wurm]
			
			#if no entropy is present in the chunks, entropytest doesn't make sense
			if (np.sum((entro*temp_areaImg)) == 0) or (temp_size == 0):
				#print(">>> not enough entropy for score")
				breakentropy = 1
				break
			else:
				relative_entropy = np.sum((entro*temp_areaImg))/temp_size
			
			#scored_entropy = (math.log(temp_size)/relative_entropy)*temp_size #logarithm entropy awesomeness! --> plot machen
			scored_entropy = math.log(relative_entropy)*temp_size
			
			#print("entropy:",i,":",relative_entropy,"size:",temp_size,"score:",scored_entropy,"total:",np.sum((entro*temp_areaImg)))

			bestobj[i-1] = scored_entropy
	else:
		breakentropy = 1
	
	#------ now choose the chunk with the lowest entropy to be best_******
	if (breakentropy == 0):
		if (allareas.shape[0] >= 3):
			n_largest = allareas.argsort()[-(argmin(bestobj)+1)] #take the chunk with the lowest entropy
		for i in range(0,label_wurm.max() + 1): # in dem 1D array area werden nun alle objekte die nicht das n-größte sind gleich null gesetzt, so das nur das größte übrig bleibt
			if i != n_largest:
				allareas[i] = 0
			
		best_size = max(allareas)
	
	#----- finish the mask
	area[argmax(area)] = 1 # dem verbleibenden objekt anstatt seiner pixelsumme an jeder position, den wert 1 zuweisen.
	areaImg = area[label_wurm] #area ist ein (1d)array mit den größen für jedes objekt. diese größenwerte werden nun den objekten in label_wurm zugewiesen, also allen bis auf das größte objekt der wert null. areaIMG enthält an allen positionen des verbleinden objekts den summenwert des gesamten objekts. könnte auch einfach 1 sein weil sowieso nurnoch eins übrig ist 
	
	#if entropy didn't work use the classic values from size
	if (breakentropy == 1):
		best_areaImg = areaImg
		best_size = size
	else:
		allareas[argmax(allareas)] = 1
		best_areaImg = allareas[label_wurm]
	
	return(areaImg,size,best_areaImg,best_size)

def wurmcal(start_frame,stop_frame,pcrop,directory,targetsize,bestedge,chunknumber,Globalstart,total_chunks):
	#usage of try to print errors outside of multicore
	try:
		#initialize stuff
		outfile = open(directory + "/Tracking_data/data%.4i.txt" % start_frame,"w")
		
		start = time.time()
		cframe = start_frame
		total_frames = stop_frame - start_frame
		
		#Start main program loop
		while cframe < stop_frame:
			###################################################################
			#--- fancy progress window ----------------------------------------
			#calculate progress
			progress = (cframe + 1 - start_frame)
			if (chunknumber + 1) <= 9:
				bar_size = 60
			else:
				bar_size = 59
			faktor = bar_size/total_frames
			
			#print progress bar for each chunk
			a = chunknumber + 2 #number of lines on top of process block
			while a > 17:
				a = a - 17
				a += 1
			
			#print("cframe",cframe,"---",chunknumber)
			
			print("Total # of chunks:",total_chunks,"time:","%.1f" %((time.time()-Globalstart)/60),"min.",end="\r")
			print("\033[1B","---","\033[1A",end="\r",sep="") #go down and then up one line
			print("\033[%aB" %a,"Chunk #",chunknumber+1,":","[","="*math.ceil(progress*faktor),"."*(bar_size-(math.ceil(progress*faktor))),"]"," ",int(((chunknumber+1)/total_chunks)*100),"%","\033[%aA" %a,end="\r",sep="")
			
			####################################################################
			#----- Load picture and apply threshold ----------------------------
			
			wurmsource = directory + '/%05d.tif' % cframe
			wurm = scipy.misc.imread(wurmsource) #get image, call it wurm
			wurm = img_as_float(wurm) #convert image to float64
			
			#set the detected edge (from canny edge detection) to largest value in image ("white")
			white_reference = wurm.max()
			wurm[bestedge == False] = white_reference
			
			wurm = nt.crop(wurm,pcrop[0],pcrop[1],pcrop[2],pcrop[3]) #crop image according to GUI input
			
			#allow for worm growth
			first_targetsize = 5500
			last_targetsize = 5500
			#growth_rate = (last_targetsize - first_targetsize)/920
			growth_rate=0
			
			#Get threshhold
			old_areaImg,old_size,areaImg,size = getthreshold(wurm,targetsize + (growth_rate * cframe))
			
			####################################################################
			#------ Closing holes, adding border to avoid stickiness -----------
			
			closedwurm = nt.addbordern(areaImg,30)
			closedwurm = binary_closing(closedwurm,disk(9),out=None)
			closedwurm = nt.removebordern(closedwurm,30)
			
			com = ndimage.center_of_mass(closedwurm)
			
			#----- Black/White worm
			color_COM = wurm.copy()
			
			color_COM[com[0],com[1]] = white_reference
			color_COM[(com[0]) - 3:(com[0]) + 3,(com[1]) - 3:(com[1]) + 3] = white_reference
			
			####################################################################
			#----- Output Data -------------------------------------------------
			wurmout = directory + '/Centroid_pictures/out_CT_%05d.tif' % cframe
			scipy.misc.imsave(wurmout,color_COM)
			
			outfile.write(str(cframe) + "," + str(com[0]) + "," + str(com[1]) + "\n")
			
			####################################################################
			#----- finish it up
			cframe += 1
		
		outfile.close()
		stop = time.time()
	
	except Exception as e:
		print(e)
		print(">>> error in frame:",cframe)

def wurmstart(startframe,endframe,pcrop,directory,blocksize,framecount,targetsize,bestedge,Globalstart,qposition,c_qposition,px_size,frame_time):
	
	global total_chunks
	#---set blocksize
	global block
	global blockchange
	
	#--- override from GUI
	blocksize = ntc.CT_blocksize
	
	block = list(nt.xfrange(1,framecount + 1,blocksize)) #list where to create a new mulitprocessing block at least
	blockchange = list(nt.xfrange(1,framecount + 1,blocksize)) #list where burst mode ends
	
	#---create data directories
	try:
		os.stat(directory + "/Tracking_data")
	except:
		os.mkdir(directory + "/Tracking_data")
	try:
		os.stat(directory + "/Centroid_pictures")
	except:
		os.mkdir(directory + "/Centroid_pictures")
	
	#---timing stuff
	allstart = time.time()
	duration = (endframe - startframe)/ntc.ct_currentfpm
	qprogress = c_qposition - 1
	qsize = 54
	qfactor = qsize/(qposition -1)
	
	print("-----------------------------------------------------------------------")
	print("Total Progress:","[","#"*math.ceil(qprogress*qfactor),"-"*(qsize-math.ceil(qprogress*qfactor)),"]",sep="")
	print("Estimated Block Time:","%.1f" % duration,"min.")
	print(directory)
	print("-----------------------------------------------------------------------")
	
	#------ Main program ---------------------------
	global pool
	pool = Pool(processes=ntc.ct_parallel_processes)
	
	total_chunks = multicorewurm(startframe,endframe,blockchange,pcrop,directory,targetsize,bestedge,Globalstart)
	
	pool.close()
	pool.join()
	#-----------------------------------------------
	
	#---timing stuff
	allend = time.time()
	
	a = total_chunks + 2 #move cursor down number of chunks + empty line
	
	print("\n"*a,sep="")
	print("---------------------------------------------------------------")
	print("block time elapsed:","%.2f" % ((allend - allstart)/60),"min.")
	fpm = ((endframe - startframe)/(allend - allstart))*60
	print("pictures per minute:","%.1f" % fpm)
	
	#----- merge all output files into single data file and draw informative graphs
	wormname = directory.split('/')[-2].replace(" ","_")
	Graphlabel = "Centroid Speed " + directory.split('/')[-2] #Get Name of Folder(worm)
	
	filenames = sorted(glob.glob(directory + '/Tracking_data/*.txt'))
	with open(directory + '/Tracking_data/' + wormname + '_Centroid.dat', 'w') as mergefile:
		mergefile.write("#" + str(Graphlabel) + "\n")
		mergefile.write("#Frame No.,Centroid X,Centroid Y\n")
		for fname in filenames:
			with open(fname) as infile:
				for line in infile:
					mergefile.write(line)
	
	#---delete txt files
	filelist = glob.glob(directory + "/Tracking_data/*.txt")
	for f in filelist:
		os.remove(f)
	
	#---delte working pictures
	filelist = glob.glob(directory + "/*.tif")
	for f in filelist:
		os.remove(f)
	
	#---save figure
	datafile = np.loadtxt(directory + '/Tracking_data/' + wormname + '_Centroid.dat',delimiter=",") #load datafile into numpy array
	movedata = np.zeros(((len(datafile[:,0]),2)))
	movedata[:,0] = datafile[:,0]
	
	for i in range(0,len(datafile[:,0])-1):	#calculate moving distance
		a = datafile[i,1] - datafile[i+1,1]
		b = datafile[i,2] - datafile[i+1,2]
		movedata[i,1] = hypot(a,b) * px_size / frame_time
	
	#add speed to the table
	alldata = np.zeros((datafile.shape[0],datafile.shape[1]+1))
	for i in range(0,len(datafile[:,0])):
		for j in range(0,datafile.shape[1]):
			alldata[i,j] = datafile[i,j]
		alldata[i,datafile.shape[1]] = movedata[i,1]

	#save table with calculated speed values
	np.savetxt(directory + '/Tracking_data/' + wormname + '_Centroid.dat',alldata,delimiter=',',header="Centroid Speed " + wormname + "\nFrame No.,Centroid X,Centroid Y,Centroid Speed")

	#try to calculate mean values
	try:
		block1mean,block2mean,block3mean = nt.blockmean(ntc.block1,ntc.block2,ntc.block3,movedata,1)
		print("Average moving speed",block1mean,block2mean,block3mean)
		meanValues = "Block 1: " + str(block1mean) + "\n" + "Block 2: " + str(block2mean) + "\n" + "Block 3: " + str(block3mean)
	except:
		meanValues = "not enough data\nfor mean values"
	
	#draw and save graph
	plt.title(Graphlabel)
	plt.xlabel('Frame #')
	plt.xticks(np.arange(min(movedata[:,0]),max(movedata[:,0]),40)-1)
	
	plt.ylabel('Speed in $\mu$m/s')
	plt.ylim((ntc.y_minimum,ntc.y_maximum))
	
	plt.annotate(meanValues,xy=(0.05,0.9), xycoords='axes fraction', xytext=(0,0), textcoords='offset points',ha='left', va='center')
	
	plt.plot(movedata[:,0],movedata[:,1], 'g,', label="speed")
	savefig(directory + '/Tracking_data/' + wormname + '_CT_centroid_speed.pdf', bbox_inches='tight')
	plt.close()
	
	#save datafile and graphs to shared folder
	if ntc.shared_folder == 1:
		try:
			os.stat(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		except:
			os.mkdir(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		
		shutil.copy2(directory + '/Tracking_data/' + wormname + '_Centroid.dat',os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data/")
		
	if ntc.shared_plots == 1:
		try:
			os.stat(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		except:
			os.mkdir(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		
		shutil.copy2(directory + '/Tracking_data/' + wormname + '_CT_centroid_speed.pdf',os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data/")

	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
	
