#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
#  pytest.py
#  
#  Copyright 2014 Unknown <birk@antergos>
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
#  MA 02110-1301, USA.

import scipy
from scipy import ndimage, interpolate
from scipy.interpolate import interp1d

import numpy as np
import matplotlib.pyplot as plt

from pylab import *
import mahotas

from skimage import morphology, filters, measure, exposure, img_as_float
from skimage.morphology import binary_closing, disk
from skimage.filters.rank import entropy
from skimage.measure import perimeter
from skimage.restoration import denoise_bilateral

import time
import sys
from multiprocessing import Process,Pool

import resource

import glob
import os
import shutil
import random
import math

import warnings

import nosetracker_config as ntc

#----- Config
silent = ntc.silent
showplots = ntc.showplots

timings = 0
warnings.filterwarnings("ignore")

#----- Variables
spalten = 4
zeilen = 2

#----- Print Python Version
print("Python Version:",sys.version)

def blockmean(block1_size,block2_size,block3_size,frames,frame_column):
	#calculate mean along the chosen blocks
	block1mean = []
	block2mean = []
	block3mean = []
	
	stacksize = block1_size + block2_size + block3_size
	
	#block1
	start_block1 = np.arange(min(frames[:,0]) - 1 ,max(frames[:,0]),stacksize)
	stop_block1 = np.arange(min(frames[:,0]) - 1 + block1_size,max(frames[:,0]) + 1,stacksize)
	
	#block2
	start_block2 = np.arange(min(frames[:,0]) - 1 + block1_size,max(frames[:,0]),stacksize)
	stop_block2 = np.arange(min(frames[:,0]) - 1 + block1_size + block2_size,max(frames[:,0]) + 1,stacksize)
	
	#block3
	start_block3 = np.arange(min(frames[:,0]) - 1 + block1_size + block2_size,max(frames[:,0]),stacksize)
	stop_block3 = np.arange(min(frames[:,0]) - 1 + block1_size + block2_size + block3_size,max(frames[:,0]) + 1,stacksize)
	
	for i in range(0,len(stop_block1)):
		block1mean.append(frames[start_block1[i]:stop_block1[i],frame_column].mean())
		
	for i in range(0,len(stop_block2)):
		block2mean.append(frames[start_block2[i]:stop_block2[i],frame_column].mean())

	for i in range(0,len(stop_block3)):
		block3mean.append(frames[start_block3[i]:stop_block3[i],frame_column].mean())
	
	block1mean = sum(block1mean)/len(block1mean)
	block2mean = sum(block2mean)/len(block2mean)
	block3mean = sum(block3mean)/len(block3mean)
	
	return(block1mean,block2mean,block3mean)
	
#~ def avgbrightness(nparray):
	#~ #gets the average brightness of an image in 8bit
	#~ avgbrg = sum(nparray)/(nparray.shape[0]*nparray.shape[1])
	#~ return(avgbrg)

def multicorewurm(startf,endf,blockc,pcrop,directory,targetsize,bestedge,startnoseposition,Globalstart):
	#---generate list of parallel processable chunks
	endf = endf + 1 #to generate the correct blocks (last frame specifically)
	chunks = [startf]
	
	#generate new blocks, if the range is bigger than a single block
	try:
		#find the end of the first blockend after start if possible
		for i in range(0,len(blockc)):
			if ((startf < blockc[i]) and (blockc[i] < endf)):
				chunks.append(blockc[i])
				index = i
				break
		
		#genereate successive blocks from blocklist
		for i in range(index,len(blockc)-1):
			if (blockc[i+1] < endf):
				chunks.append(blockc[i])
				chunks.append(blockc[i+1])
			else:
				break
			
		#generate last chunk from blocklist till end
		chunks.append(chunks[-1])
		chunks.append(endf)

	except:
		chunks.append(endf)
	
	total_chunks = int(len(chunks)/2)
	
	#spawn as much processes as blocks are available or until endframe is reached
	for i in range(0,int(len(chunks)/2)):
		blockstart = chunks[i*2]
		blockend = chunks[i*2+1]
		
		chunknumber = i
		
		pool.apply_async(wurmcal, args=(blockstart,blockend,pcrop,directory,targetsize,bestedge,startnoseposition,chunknumber,Globalstart,total_chunks))
		#print("chunk enqued, block:",blockstart,"-",blockend - 1)
	
	return(total_chunks)

def crop(nparray,oben,rechts,unten,links):
	#cropping the source image
	unten = nparray.shape[0] - unten
	rechts = nparray.shape[1] - rechts
	
	vertical = oben + nparray.shape[0] - unten
	horizontal = links + nparray.shape[1] - rechts
	
	outarray = np.zeros((nparray.shape[0]-vertical,nparray.shape[1]-horizontal))
	outarray[:,:] = nparray[oben:unten,links:rechts]
	
	return(outarray)

def amask(index,radius,array):
	#create a circular mask and return pixelsum within radius
	#t@10k=8.44
	a,b = index
	nx,ny = array.shape
	y,x = np.ogrid[-a:nx-a,-b:ny-b]
	mask = x*x + y*y <= radius*radius
	
	return(sum(array[mask]),sum(mask)) #returns size of the mask as well

def xfrange(start, stop, step):
	#function to create floating steps
    while start < stop:
        yield start
        start += step

def addbordern(nparray,n):
	outarray = np.zeros((nparray.shape[0]+n*2,nparray.shape[1]+n*2))
	outarray[n:-n,n:-n] = nparray[:,:]
	return(outarray)
	
def removebordern(nparray,n):
	outarray = np.zeros((nparray.shape[0]-(2*n),nparray.shape[1]-(2*n)))
	outarray[:,:] = nparray[n:-n,n:-n]
	return(outarray)

def HMlineends(linearray):
	#------ hit and miss algorithm to find lineends ---------------------
	
	#define pattern matching arrays for each possible direction
	Sline = np.array([[0,0,0],[0,1,0],[0,1,0]])
	Nline = np.array([[0,1,0],[0,1,0],[0,0,0]])
	Wline = np.array([[0,0,0],[1,1,0],[0,0,0]])
	Eline = np.array([[0,0,0],[0,1,1],[0,0,0]])
	SWline = np.array([[0,0,0],[0,1,0],[1,0,0]])
	SEline = np.array([[0,0,0],[0,1,0],[0,0,1]])
	NWline = np.array([[1,0,0],[0,1,0],[0,0,0]])
	NEline = np.array([[0,0,1],[0,1,0],[0,0,0]])
	
	#find linends
	N = ndimage.binary_hit_or_miss(linearray,structure1=Nline)
	S = ndimage.binary_hit_or_miss(linearray,structure1=Sline)
	E = ndimage.binary_hit_or_miss(linearray,structure1=Eline)
	W = ndimage.binary_hit_or_miss(linearray,structure1=Wline)
	NW = ndimage.binary_hit_or_miss(linearray,structure1=NWline)
	NE = ndimage.binary_hit_or_miss(linearray,structure1=NEline)
	SW = ndimage.binary_hit_or_miss(linearray,structure1=SWline)
	SE = ndimage.binary_hit_or_miss(linearray,structure1=SEline)
	
	HMarray = S+N+E+W+NE+NW+SE+SW
	HMcoords = np.transpose(np.nonzero(HMarray)) #finde alle punkte != 0 und schreibe koordinaten in 2D array (punkte sind die lieends)
	
	return (HMarray,HMcoords)

def HMlineends8(linearray):
	#------ hit and miss algorithm to find lineends ---------------------
	
	#define pattern matching arrays for each possible direction
	Sline = np.array([[0,0,0],[0,1,0],[0,1,0]])
	Nline = np.array([[0,1,0],[0,1,0],[0,0,0]])
	Wline = np.array([[0,0,0],[1,1,0],[0,0,0]])
	Eline = np.array([[0,0,0],[0,1,1],[0,0,0]])
	SWline = np.array([[0,0,0],[0,1,0],[1,0,0]])
	SEline = np.array([[0,0,0],[0,1,0],[0,0,1]])
	NWline = np.array([[1,0,0],[0,1,0],[0,0,0]])
	NEline = np.array([[0,0,1],[0,1,0],[0,0,0]])
	
	NWNline = np.array([[1,0,0],[1,1,0],[0,0,0]])
	NOOline = np.array([[0,0,1],[0,1,1],[0,0,0]])
	NWWline = np.array([[1,1,0],[0,1,0],[0,0,0]])
	NNOline = np.array([[0,1,1],[0,1,0],[0,0,0]])
	SSOline = np.array([[0,0,0],[0,1,0],[0,1,1]])
	SWSline = np.array([[0,0,0],[0,1,0],[1,1,0]])
	SWWline = np.array([[0,0,0],[1,1,0],[1,0,0]])
	SOOline = np.array([[0,0,0],[0,1,1],[0,0,1]])
	
	#find linends
	N = ndimage.binary_hit_or_miss(linearray,structure1=Nline)
	S = ndimage.binary_hit_or_miss(linearray,structure1=Sline)
	E = ndimage.binary_hit_or_miss(linearray,structure1=Eline)
	W = ndimage.binary_hit_or_miss(linearray,structure1=Wline)
	NW = ndimage.binary_hit_or_miss(linearray,structure1=NWline)
	NE = ndimage.binary_hit_or_miss(linearray,structure1=NEline)
	SW = ndimage.binary_hit_or_miss(linearray,structure1=SWline)
	SE = ndimage.binary_hit_or_miss(linearray,structure1=SEline)
	
	NWN = ndimage.binary_hit_or_miss(linearray,structure1=NWNline)
	NOO = ndimage.binary_hit_or_miss(linearray,structure1=NOOline)
	NWW = ndimage.binary_hit_or_miss(linearray,structure1=NWWline)
	NNO = ndimage.binary_hit_or_miss(linearray,structure1=NNOline)
	SSO = ndimage.binary_hit_or_miss(linearray,structure1=SSOline)
	SWS = ndimage.binary_hit_or_miss(linearray,structure1=SWSline)
	SWW = ndimage.binary_hit_or_miss(linearray,structure1=SWWline)
	SOO = ndimage.binary_hit_or_miss(linearray,structure1=SOOline)

	HMarray = S+N+E+W+NE+NW+SE+SW+NWN+NOO+NWW+NNO+SSO+SWS+SWW+SOO
	HMcoords = np.transpose(np.nonzero(HMarray)) #finde alle punkte != 0 und schreibe koordinaten in 2D array (punkte sind die lineends)
	
	return (HMarray,HMcoords)

def HMlinejunctions(linearray):
	#------ hit or miss algorithm to find linejunctions ----------------
	
	#define pattern matching arrays for 3-connected Y/T lines and 4-connectedness
	SYjunction = np.array([[1,0,1],[0,1,0],[0,1,0]])
	WYjunction = np.array([[0,0,1],[1,1,0],[0,0,1]])
	NYjunction = np.array([[0,1,0],[0,1,0],[1,0,1]])
	EYjunction = np.array([[1,0,0],[0,1,1],[1,0,0]])
	SWYjunction = np.array([[0,1,0],[0,1,1],[1,0,0]])
	NWYjunction = np.array([[1,0,0],[0,1,1],[0,1,0]])
	NEYjunction = np.array([[0,0,1],[1,1,0],[0,1,0]])
	SEYjunction = np.array([[0,1,0],[1,1,0],[0,0,1]])
	
	NTjunction = np.array([[0,1,0],[1,1,1],[0,0,0]])
	ETjunction = np.array([[0,1,0],[0,1,1],[0,1,0]])
	STjunction = np.array([[0,0,0],[1,1,1],[0,1,0]])
	WTjunction = np.array([[0,1,0],[1,1,0],[0,1,0]])
	SWTjunction = np.array([[1,0,0],[0,1,0],[1,0,1]])
	NWTjunction = np.array([[1,0,1],[0,1,0],[1,0,0]])
	NETjunction = np.array([[1,0,1],[0,1,0],[0,0,1]])
	SETjunction = np.array([[0,0,1],[0,1,0],[1,0,1]])
	
	NXjunction = np.array([[0,1,0],[1,1,1],[0,1,0]])
	NEXjunction = np.array([[1,0,1],[0,1,0],[1,0,1]])
	
	#find junctions in skeleton
	SY = ndimage.binary_hit_or_miss(linearray,structure1=SYjunction)
	WY = ndimage.binary_hit_or_miss(linearray,structure1=WYjunction)
	NY = ndimage.binary_hit_or_miss(linearray,structure1=NYjunction)
	EY = ndimage.binary_hit_or_miss(linearray,structure1=EYjunction)
	SWY = ndimage.binary_hit_or_miss(linearray,structure1=SWYjunction)
	NWY = ndimage.binary_hit_or_miss(linearray,structure1=NWYjunction)
	NEY = ndimage.binary_hit_or_miss(linearray,structure1=NEYjunction)
	SEY = ndimage.binary_hit_or_miss(linearray,structure1=SEYjunction)
	
	NT = ndimage.binary_hit_or_miss(linearray,structure1=NTjunction)
	ET = ndimage.binary_hit_or_miss(linearray,structure1=ETjunction)
	ST = ndimage.binary_hit_or_miss(linearray,structure1=STjunction)
	WT = ndimage.binary_hit_or_miss(linearray,structure1=WTjunction)
	SWT = ndimage.binary_hit_or_miss(linearray,structure1=SWTjunction)
	NWT = ndimage.binary_hit_or_miss(linearray,structure1=NWTjunction)
	NET = ndimage.binary_hit_or_miss(linearray,structure1=NETjunction)
	SET = ndimage.binary_hit_or_miss(linearray,structure1=SETjunction)
	
	NX = ndimage.binary_hit_or_miss(linearray,structure1=NXjunction)
	NEX = ndimage.binary_hit_or_miss(linearray,structure1=NEXjunction)
	
	
	HMarray = SY+NY+EY+WY+NEY+NWY+SEY+SWY+NT+ET+ST+WT+SWT+NWT+NET+SET+NX+NEX
	HMcoords = np.transpose(np.nonzero(HMarray)) #finde alle punkte != 0 und schreibe koordinaten in 2D array (punkte sind die junctions)
	
	return (HMarray,HMcoords)

def getthreshold (nparray,targetsize):
	#--- apply blurring filter
	
	if ntc.blur_method == 0:
		wurmf = ndimage.gaussian_filter(nparray,ntc.gauss_radius) #gaussfilterradius, 1.5 seems to be ok
	if ntc.blur_method == 1:
		wurmf = denoise_bilateral(nparray, sigma_range=ntc.bilateral_range, sigma_spatial=ntc.bilateral_spatial)
	
	#set the thresholds for contrast to brightest/darkest pixel
	startthresh = wurmf.min()
	endthresh = wurmf.max()
	
	#initialize stuff
	size = 0
#	tired = 0
	supertired = 0
	breakentropy = 0 #check to avoid second comparison if a zero was encountered in the first
	c = (random.random() * (endthresh-startthresh)) + startthresh #still better than zero
	bestobj = np.zeros((3))
		
	###### monte carlo approach #####
	while (not(targetsize * ntc.nt_upper_size_limit < size < targetsize * ntc.nt_upper_size_limit)): #size should be within 0.5% of target size
		wurmfertig = wurmf < c
		#--- label all objects
		label_wurm, num = ndimage.label(wurmfertig) #label alle objekte die nach dem threshold noch übrig sind
		
		#------ calculate areas, delete all but biggest --> könnte noch optimiert werden
		area = ndimage.sum(wurmfertig, label_wurm, index=arange(label_wurm.max() + 1)) #sum addiert alle zu einem label gehörenden positionen (binäre 1en) zusammen. wurmfertig ist binäres array, label_wurm enthält die label der areas in wurmfertig. index bestimmt über welche bereiche des arrays die summen gebildet werden sollen. in diesem fall die summe des binären arrays wurmfertig mit den Bereichen die gelabled wurden. und das von allen labeln (also label.max+1)
		
		#------ keep unaltered area-table --> optimize to do this only on last run(take latest threshold etc.)
		allareas = area.copy()
		
		#------ take biggest chunk
		largest = argmax(area)
		for i in range(0,label_wurm.max() + 1): # in dem 1D array area werden nun alle objekte die nicht das größte sind gelich null gesetzt, so das nur das größte übrig bleibt
			if i != argmax(area):
				area[i] = 0
		
		size = max(area)
		
		#------ reduce range wherein random numbers are looked for
		if size > targetsize:
			endthresh = c
		if size < targetsize:
			startthresh = c
		
		c = (random.random() * (endthresh-startthresh)) + startthresh		
		
		#--- prevent algorithm from getting stuck if threshold too narrow
#		tired += 1
		supertired += 1
		
#		#check if the upper and lower threshold limit is sufficiently close to each other
#		if (endthresh - startthresh) > 0.00001:
#			tired = 0
#		if tired == 3:
#			break
		
		if supertired >= ntc.max_attempts:
			break
		
	#------ determine chunk with lowest entropy (only works if at least 3 chunks)
	entro = entropy(nparray, disk(3))
	
	if (allareas.shape[0] >= 3): #check against number of objects
		for i in range(1,4): #check the 3 largest objects (not 0-3 because backwards)
			
			temp_allareas = allareas.copy() #make a copy of the area-array so we can modify it for entropy calc
			n_largest = temp_allareas.argsort()[-i] #gets the index of the n-largest([-i]-largest) object
			
			for j in range(0,label_wurm.max() + 1): #remove all but chosen object
				if j != n_largest:
					temp_allareas[j] = 0
			
			temp_size = max(temp_allareas)
			temp_allareas[argmax(temp_allareas)] = 1
			temp_areaImg = temp_allareas[label_wurm]
			
			#if no entropy is present in the chunks, entropytest doesn't make sense
			if (np.sum((entro*temp_areaImg)) == 0) or (temp_size == 0):
				#print(">>> not enough entropy for score")
				breakentropy = 1
				break
			else:
				relative_entropy = np.sum((entro*temp_areaImg))/temp_size
			
			#scored_entropy = (math.log(temp_size)/relative_entropy)*temp_size #logarithm entropy awesomeness! --> plot machen
			scored_entropy = math.log(relative_entropy)*temp_size
			
			#print("entropy:",i,":",relative_entropy,"size:",temp_size,"score:",scored_entropy,"total:",np.sum((entro*temp_areaImg)))

			bestobj[i-1] = scored_entropy
	else:
		breakentropy = 1
	
	#------ now choose the chunk with the lowest entropy to be best_******
	if (breakentropy == 0):
		if (allareas.shape[0] >= 3):
			n_largest = allareas.argsort()[-(argmin(bestobj)+1)] #take the chunk with the lowest entropy
		for i in range(0,label_wurm.max() + 1): # in dem 1D array area werden nun alle objekte die nicht das n-größte sind gleich null gesetzt, so das nur das größte übrig bleibt
			if i != n_largest:
				allareas[i] = 0
			
		best_size = max(allareas)
	
	#----- finish the mask
	area[argmax(area)] = 1 # dem verbleibenden objekt anstatt seiner pixelsumme an jeder position, den wert 1 zuweisen.
	areaImg = area[label_wurm] #area ist ein (1d)array mit den größen für jedes objekt. diese größenwerte werden nun den objekten in label_wurm zugewiesen, also allen bis auf das größte objekt der wert null. areaIMG enthält an allen positionen des verbleinden objekts den summenwert des gesamten objekts. könnte auch einfach 1 sein weil sowieso nurnoch eins übrig ist 
	
	#if entropy didn't work use the classic values from size
	if (breakentropy == 1):
		best_areaImg = areaImg
		best_size = size
	else:
		allareas[argmax(allareas)] = 1
		best_areaImg = allareas[label_wurm]
	
	return(areaImg,size,best_areaImg,best_size)

def nosefinder(wurm,wormcore,offsetx,offsety,closedwurm,newblock,lastnose):
	#initialize stuff
	endpoints = []
	best_end = np.array([0,0,0,0])
	consensus1 = [0,0,0]
	
	####################################################################
    #------ look at both calculated ends -------------------------------	
	for flipnose in range(0,2):

		####################################################################
		#------ find (correct) lineends ------------------------------------
		skeletoncoord = np.transpose(np.nonzero(wormcore))
		
		#------ define arrays
		core_array,core_ends = HMlineends(wormcore) #get lineends
		distmatrix = scipy.spatial.distance.cdist(core_ends,core_ends) #creates a matrix of distances between each point
		
		long1 = np.where(distmatrix==distmatrix.max()) #gets tuple with both indices ((a,b)(b,a))
		long2 = long1[0] #save first tuple

		#initialize
		end1 = np.zeros((2))
		end2 = np.zeros((2))
		
		end1 = core_ends[long2[0]]
		end2 = core_ends[long2[1]]
		
		#arbitrary assignment
		nose = end1
		tail = end2
		
		#change which end is assumed tail/nose for next run 
		if flipnose == 0:
			x = nose[0]
			y = nose[1]
			nose[0] = tail[0]
			nose[1] = tail[1]
			tail[0] = x
			tail[1] = y
			flipnose = 0
		
		#------ delete nose-point from skeletoncoord list
		for i in range (0,len(skeletoncoord)):
			if all(skeletoncoord[i,:] == (nose)):
				skeletoncoord = delete(skeletoncoord, i, 0) #array,index,axis
				break
		
		#------ set starting point for skeleton sort
		sorted_skeleton = np.zeros((skeletoncoord.shape))

		sorted_skeleton[0,0] = nose[0]
		sorted_skeleton[0,1] = nose[1]
		
		####################################################################
		#----- sort coordinates
		for j in range(0,len(sorted_skeleton) - 1):
			d1_smallest = 10000 #arbitrary large number, init for smallest distance
			for i in range(0,len(skeletoncoord)):
				a = sorted_skeleton[j,0] - skeletoncoord[i,0]
				b = sorted_skeleton[j,1] - skeletoncoord[i,1]
				d1 = hypot(a,b)
			
				if (0 < d1 < d1_smallest):
					index = i
					d1_smallest = d1
				if (d1_smallest < 1.5): #break for loop if an 8-connected pixel(1 or 1.414..) is found (wont get closer)
					break

			sorted_skeleton[j+1,0] = skeletoncoord[index,0]
			sorted_skeleton[j+1,1] = skeletoncoord[index,1]
			skeletoncoord = delete(skeletoncoord, index, 0)
		
		####################################################################
		#------ extrapolate worm -------------------------------------------

		#------ process coordinates
		sorted_skeleton = sorted_skeleton[:(sorted_skeleton.shape[0]*0.75):1] #die ersten 75% koordinaten,jede
		
		skeletoncoordx = sorted_skeleton[:,0] #column 0
		skeletoncoordy = sorted_skeleton[:,1] #column 1
		
		x = skeletoncoordy
		y = - skeletoncoordx + offsety
		
		#------ fit a spline to the coordinates, x and y axis are interpolated towards t
		t = np.arange(x.shape[0], dtype=float) #t ist anzahl an werten
		t /= t[-1] #t ist nun unterteilt in 0 bis 1
		nt = np.linspace(0, 1, 100) #nt ist array mit werten von 0 bis 1 mit 100 zwischenwerten
		
		x1 = scipy.interpolate.spline(t, x, nt) #The x values where spline should estimate the y values
		y1 = scipy.interpolate.spline(t, y, nt)
		
		#------ create a new linear space for nnt in which an extrapolation from the interpolated spline will be made 
		nnt = np.linspace(-1, 1, 100) #values <0 are extrapolated (interpolation started at the nose(=0)

		x1fit = np.polyfit(nt,x1,2) #fits a polynomial function of the nth order with the spline as input, output are the function parameters
		y1fit = np.polyfit(nt,y1,2)
		
		xpoly = np.poly1d(x1fit) #genereates the function based on the parameters obtained by polyfit
		ypoly = np.poly1d(y1fit)
		
		#------ write coordinates into array
		extrapolatedworm = np.zeros(shape=(len(nnt),2))
		extrapolatedworm[:,0] = xpoly(nnt)
		extrapolatedworm[:,1] = ypoly(nnt)
		
		#~ tempwurm = wurm.copy()
		
		#~ print(extrapolatedworm)
		
		#~ for i in range(0,len(extrapolatedworm[:,0])):
			#~ if (int(extrapolatedworm[i,0]) < len(tempwurm[0])):
				#~ if (int(extrapolatedworm[i,1]) < len(tempwurm[:,0])):
					#~ tempwurm[int(extrapolatedworm[i,0]),int(extrapolatedworm[i,1])] = 255
		
		
		####################################################################
		#------ match function to nosetip ----------------------------------
		
		nosesegment = np.zeros((closedwurm.shape))
		for i in xfrange (-0.5,1,0.001):		#distance optimieren(-0.5,1,0.001)
			if ((0 <= (xpoly(i)) <= (offsety)) and (0 <= (-ypoly(i)+offsety) <= (offsetx))):
				if (closedwurm[(-ypoly(i)+offsety),(xpoly(i))] >= 1):
					nosesegment[(-ypoly(i)+offsety),(xpoly(i))] = 5
		
		tiparray,tipcoords = HMlineends8(nosesegment)
		
		##if flipnose == 1:
			##scipy.misc.imsave("/home/purzel/Programming/Nosetracker/5_nosesegmentf1.tif",nosesegment)
		##if flipnose == 0:
			##scipy.misc.imsave("/home/purzel/Programming/Nosetracker/5_nosesegmentf2.tif",nosesegment)
		
		#------ if resulting nosesegment has a weird shape (not 2 ends), thin it, probably artefact from function
		if (len(tipcoords) < 2): #len() returns rows
			nosesegment = morphology.skeletonize(nosesegment > 0)
			tiparray,tipcoords = HMlineends(nosesegment)
			#print(">>> weird spline")
		
		#------ if no tips, skip frame
		if (len(tipcoords) == 0):
			tipcoords = np.zeros((2,2))
			#print(">>> frame skipped!")
			
		#------ if there is still only one lineend, we probably got a sigma-like circle
		#-nothing sensible to do here, copy single end to both coordinates
		if (len(tipcoords) == 1):
			temparray = np.zeros((2,2))
			temparray[0] = tipcoords
			temparray[1] = tipcoords
			tipcoords = np.zeros((2,2))
			tipcoords[0] = temparray[0]
			tipcoords[1] = temparray[1]
			#print(">>> spline-end < 2")
			
		#------ check all lineends, whichever is closer to the function=0, this is the function origin 
		#=== check if funczero or nose is ok
		funczero = np.zeros((2))
		funczero[0] = (-ypoly(0)+offsety)
		funczero[1] = (xpoly(0))
		
		#------ check for the best information on the last noseposition
		#check against the presumed nose taken from the skeleton (function at 0)
		distfunc_old = 100000
		for i in range(0,tipcoords.shape[0]):
			a = tipcoords[i,0] - funczero[0]
			b = tipcoords[i,1] - funczero[1]
			distfunc = hypot(a,b)
			if (distfunc < distfunc_old):
				distfunc_old = distfunc
				index = i
		
		#realnose is the point where the function intersects with the worm-shape
		realnose = np.zeros((2))
		realnose = tipcoords[index]
		
		#add both "noses" to the endpoints list
		endpoints.append(realnose[0])
		endpoints.append(realnose[1])
	
	####################################################################
	#----- calculate different factors to determine correct end

	#endoints is realnose 0,1 and the second realnose 2,3
	#calculate middle distance between both points
	a = endpoints[0] - endpoints[2]
	b = endpoints[1] - endpoints[3]
	middle_distance = (hypot(a,b))/2

	#------ entropy ------
	entropywurm = entropy(wurm, disk(3))
	
	#sum of entropy within radius and relative entropy
	entropy_mass1,masksize1 = amask((endpoints[0],endpoints[1]),middle_distance * 1,entropywurm)
	entropy_mass2,masksize2 = amask((endpoints[2],endpoints[3]),middle_distance * 1,entropywurm)
	
	rel_entropy1 = entropy_mass1/masksize1
	rel_entropy2 = entropy_mass2/masksize2
	
	dif_entropy = (abs((rel_entropy1)-(rel_entropy2))/((rel_entropy1 + rel_entropy2)/2))*100
	
	#------ pixel mass ------
	#sum of pixels within radius
	pixel_mass1,pmasksize1 = amask((endpoints[0],endpoints[1]),middle_distance * 1,closedwurm)
	pixel_mass2,pmasksize2 = amask((endpoints[2],endpoints[3]),middle_distance * 1,closedwurm)
	
	dif_pixel_mass = (abs((pixel_mass1)-(pixel_mass2))/((pixel_mass1 + pixel_mass2)/2))*100
	
	#------ brightness ------
	#sum of brightness values within radius and relative brightness
	bright_mass1,bmasksize1 = amask((endpoints[0],endpoints[1]),middle_distance * 1,wurm)
	bright_mass2,bmasksize2 = amask((endpoints[2],endpoints[3]),middle_distance * 1,wurm)
	
	rel_bright1 = bright_mass1/bmasksize1
	rel_bright2 = bright_mass2/bmasksize2	
	
	dif_bright = (abs((rel_bright1)-(rel_bright2))/((rel_bright1 + rel_bright2)/2))*100
	#dif_bright = (abs((bright_mass1)-(bright_mass2))/((bright_mass1 + bright_mass2)/2))*100
	
	####################################################################
	#----- print stuff ------
	
	if (silent == 0):
		print("-----------------------")
		#print("entropy mass end1:",entropy_mass1/1000,"rel",rel_entropy1)
		#print("entropy mass end2:",entropy_mass2/1000,"rel",rel_entropy2)
		#print("difference in entropy:",dif_entropy,"%")
		
		if rel_entropy1 > rel_entropy2:
			print("end2 - entropy",dif_entropy,"%")
			consensus1[0] = 1
		else:
			print("end1 - entropy",dif_entropy,"%")
			consensus1[0] = 0
		
		#print("-----------------------")
		#print("pixel mass end1:",pixel_mass1)
		#print("pixel mass end2:",pixel_mass2)
		#print("difference in pixels:",dif_pixel_mass,"%")
		
		if pixel_mass1 > pixel_mass2:
			print("end1 - mass",dif_pixel_mass,"%")
			consensus1[1] = 0
		else:
			print("end2 - mass",dif_pixel_mass,"%")
			consensus1[1] = 1
			
		#print("-----------------------")
		#print("bright mass end1:",bright_mass1,"rel",rel_bright1)
		#print("bright mass end2:",bright_mass2,"rel",rel_bright2)
		#print("difference in bright:",dif_bright,"%")
		
		if rel_bright1 > rel_bright2:
			print("end2 - bright",dif_bright,"%")
			consensus1[2] = 1
		else:
			print("end1 - bright",dif_bright,"%")
			consensus1[2] = 0
		print("-----------------------")
	
	####################################################################
	#------ calculate score from all factors to decide wich end is to favor
	
	if dif_entropy > dif_bright and dif_entropy > dif_pixel_mass:
		if rel_entropy1 > rel_entropy2:
			#print("end2 - entropy wins")
			best_end[0] = endpoints[2]
			best_end[1] = endpoints[3]
			best_end[2] = endpoints[0]
			best_end[3] = endpoints[1]
		else:
			#print("end1 - entropy wins")
			best_end[0] = endpoints[0]
			best_end[1] = endpoints[1]
			best_end[2] = endpoints[2]
			best_end[3] = endpoints[3]
	
	elif dif_bright > dif_entropy and dif_bright > dif_pixel_mass:
		#if bright_mass1 > bright_mass2: #anstatt rel_bright
		if rel_bright1 > rel_bright2:
			#print("end2 - bright wins")
			best_end[0] = endpoints[2]
			best_end[1] = endpoints[3]
			best_end[2] = endpoints[0]
			best_end[3] = endpoints[1]
		else:
			#print("end1 - bright wins")
			best_end[0] = endpoints[0]
			best_end[1] = endpoints[1]
			best_end[2] = endpoints[2]
			best_end[3] = endpoints[3]
	else:
		if pixel_mass1 > pixel_mass2:
			#print("end1 - mass wins")
			best_end[0] = endpoints[2]
			best_end[1] = endpoints[3]
			best_end[2] = endpoints[0]
			best_end[3] = endpoints[1]
		else:
			#print("end2 - mass wins")
			best_end[0] = endpoints[0]
			best_end[1] = endpoints[1]
			best_end[2] = endpoints[2]
			best_end[3] = endpoints[3]
	
	#print avg deviation in difference
	a = dif_pixel_mass
	b = dif_bright
	c = dif_entropy
	
	dif_avg = (((abs(a-b)/((a+b)/2)) + (abs(a-c)/((a+c)/2)) + (abs(c-b)/((c+b)/2)))/3)*100
	
	#print(dif_avg,"% difference of the averages")
	####################################################################
	#----- if all else fails, compare to last position -----------------
	

	#if div_avg < 25% and  != 111 or 000 #maybe
	
	#vergleiche das ergebnis des scores mit dem noseflipper
	#newblock muss mit rein
	#wenn noseflipper und score uneinig, kontrolliere score
	
	#----- compare new lineends to position of old nose
	a = best_end[0] - lastnose[0]
	b = best_end[1] - lastnose[1]
	lastnosedist = hypot(a,b)
	
	a = best_end[2] - lastnose[0]
	b = best_end[3] - lastnose[1]
	lasttaildist = hypot(a,b)
	
	#wenn kein neuer block angefangen wurde, i.e. eine nasenposition schon bekannt ist --> manual nose
	#~ if (newblock == 0):
		#~ if (lasttaildist < lastnosedist):
			#~ #wenn der score über den 20% wert geht sollte der score das ende entscheiden
			#~ if (dif_entropy > 20) or (dif_bright > 20) or (dif_pixel_mass > 20):
				#~ print("Trust the Score!")
		#~ 
			#~ #wenn der score unter 20% ist, vertraue der letzten position
			#~ if (dif_entropy < 20) and (dif_bright < 20) and (dif_pixel_mass < 20): #flip if the other end is closer
				#~ print("Trust the last position! Nose flipped")
				#~ x = best_end[0]
				#~ y = best_end[1]
				#~ best_end[0] = best_end[2]
				#~ best_end[1] = best_end[3]
				#~ best_end[2] = x
				#~ best_end[3] = y
		#~ #wenn score und letzte position sich einig sind
		#~ else:
			#~ print("D'ACCORD")
	
	#if (newblock == 0) and (lasttaildist < lastnosedist):
	if (lasttaildist < lastnosedist):
		x = best_end[0]
		y = best_end[1]
		best_end[0] = best_end[2]
		best_end[1] = best_end[3]
		best_end[2] = x
		best_end[3] = y
	
	####################################################################
	
	return best_end[0],best_end[1],best_end[2],best_end[3]

def wurmcal(start_frame,stop_frame,pcrop,directory,targetsize,bestedge,startnoseposition,chunknumber,Globalstart,total_chunks):
	#usage of try to print errors outside of multicore
	try:
		#---globals
		global silent
		global showplots
		global timings
		global flipnose
		global f_thresh
		global lastrealnose

		outfile = open(directory + "/Tracking_data/data%.4i.txt" % start_frame,"w")
				
		#---Variables that need initialisation
		start = time.time()
		lastnose = np.array([0,0])
		lasttail = np.array([0,0])
		lastrealnose = np.array([0,0])
		firstrun = 1
		force_newblock = 0
		cframe = start_frame
		total_frames = stop_frame - start_frame
		
		################################################################
		#--- program loop
		while cframe < stop_frame:
		#for cframe in range(start_frame,stop_frame):
			####################################################################
			#------ everytime a new block is set in the blockchange list, set newblock = 1 for the next frame, also for first frame
			
			#set the manual noseposition at the specified frames
			if (cframe in startnoseposition[::3]):
				lastnose[0] = startnoseposition[((np.where(startnoseposition[::3] == cframe)[0][0])*3)+1] #returns values after the framenumber for manual nose
				lastnose[1] = startnoseposition[((np.where(startnoseposition[::3] == cframe)[0][0])*3)+2] #check every thrid value if its the same as cframe, return the index
			#toggle newblock	
			if (cframe in blockchange):
				newblock = 1
			elif (cframe == start_frame):
				newblock = 1
			elif (force_newblock == 1):
				newblock = 1
				force_newblock = 0
			else:
				newblock = 0
			
			#--- fancy progress window ---
			#calculate progress
			progress = (cframe + 1 - start_frame)
			if (chunknumber + 1) <= 9:
				bar_size = 60
			else:
				bar_size = 59
			faktor = bar_size/total_frames
			
			#print progress bar for each chunk
			a = chunknumber + 2 #number of lines on top of process block
			#if more than 16 Chunks, overwrite old ones
			while a > 17:
				a = a - 17
				a += 1
			
			print("Total # of chunks:",total_chunks,"time:","%.1f" %((time.time()-Globalstart)/60),"min.",end="\r")
			print("\033[1B","---","\033[1A",end="\r",sep="") #go down and then up one line
			print("\033[%aB" %a,"Chunk #",chunknumber+1,":","[","="*math.ceil(progress*faktor),"."*(bar_size-(math.ceil(progress*faktor))),"]"," ",int(((chunknumber+1)/total_chunks)*100),"%","\033[%aA" %a,end="\r",sep="")
			
			####################################################################
			#----- Load picture and apply threshold ----------------------------
			
			wurmsource = directory + '/%05d.tif' % cframe
			wurm = scipy.misc.imread(wurmsource) #get image, call it wurm
			wurm = img_as_float(wurm) #convert image to float64
			#set the detected edge (from canny edge detection) to largest value in image ("white")
			white_reference = wurm.max()
			wurm[bestedge == False] = white_reference
			
			wurm = crop(wurm,pcrop[0],pcrop[1],pcrop[2],pcrop[3]) #crop image according to GUI input
			
			#allow for worm growth
			first_targetsize = 5500
			last_targetsize = 5500
			#growth_rate = (last_targetsize - first_targetsize)/920
			growth_rate=0
			
			#Get threshhold
			old_areaImg,old_size,areaImg,size = getthreshold(wurm,targetsize + (growth_rate * cframe))
			
			####################################################################
			#------ Closing holes, adding border to avoid stickiness -----------
			
			##scipy.misc.imsave("/home/purzel/Programming/Nosetracker/10_areaImg.tif",areaImg)
			closedwurm = addbordern(areaImg,30)
			closedwurm = binary_closing(closedwurm,disk(9),out=None)
			
			#subplot(zeilen,spalten,4)
			#imshow(closedwurm,interpolation='none',cmap='jet')
			#title("closed holes")
			#plt.axis('off')
			
			####################################################################
			#------ skeletonization --------------------------------------------
			#skeletonwurm, mdist = morphology.medial_axis(closedwurm > 0,return_distance = True)
			if ntc.skeletonization_method == 0:
				skeletonwurm = morphology.skeletonize(closedwurm > 0) #simple skeletonization by erosion
			if ntc.skeletonization_method == 1:
				skeletonwurm = morphology.medial_axis(closedwurm > 0)
			
			if (showplots == 1):
				subplot(zeilen,spalten,5)
				imshow(skeletonwurm,interpolation='none',cmap='jet')
				title("skeletonization")
				plt.axis('off')
			##scipy.misc.imsave("/home/purzel/Programming/Nosetracker/2_Skeletonwurm.tif",skeletonwurm)

			####################################################################
			#------- removing borders ------------------------------------------
			
			closedwurm = removebordern(closedwurm,30)
			skeletonwurm = removebordern(skeletonwurm,30)
			
			offsetx = closedwurm.shape[0] - 1 #adapt for zero index
			offsety = closedwurm.shape[1] - 1
			
			#calculate center of mass
			com = ndimage.center_of_mass(closedwurm)
			
			####################################################################
			#------ reduce skeleton to segments and extract important ones -----
			
			#------ find linejunctions, label segments
			linejunctionwurm,linejunctions = HMlinejunctions(skeletonwurm)
			
			#linejunctions löschen im radius disk(n)
			linejunctionsradius = morphology.binary_dilation(linejunctionwurm,disk(2))
			
			linejunctionsradius = np.transpose(np.nonzero(linejunctionsradius))
			
			segmentedskeleton = skeletonwurm.copy()
			segmentedskeleton[linejunctionsradius[:,0],linejunctionsradius[:,1]] = 0 
			
			#einzelne segmente labeln und einfärben
			s = np.array([[1,1,1],[1,1,1],[1,1,1]]) #structure die auch diagonale verbindungen zulässt
			
			label_segmentedskeleton, segments = ndimage.label(segmentedskeleton, structure=s)
			
			#------ discard unfit segments
			#calculate size of each segment
			skeletonarea = ndimage.sum(segmentedskeleton, label_segmentedskeleton, index=arange(label_segmentedskeleton.max() + 1)) #sum addiert alle zu einem label gehörenden positionen (binäre 1en) zusammen. wurmfertig ist binäres array, label_wurm enthält die label der areas in wurmfertig. index bestimmt über welche bereiche des arrays die summen gebildet werden sollen. in diesem fall die summe des binären arrays wurmfertig mit den Bereichen die gelabled wurden. und das von allen labeln (also label.max+1)
			
			#discard elements < x (remove micro-ends)
			n = 0
			for i in range(0,len(skeletonarea)): # der reihe nach werden alle objekte auf ihre größe getestet(1D array), kleiner n werden verworfen
				if skeletonarea[i] <= 3:
					skeletonarea[i] = 0
				else:
					n += 1
					skeletonarea[i] = n
			
			segmentedskeleton = skeletonarea[label_segmentedskeleton] # alle segemente auf den korrespondierenden wert von skeletonarea setzen (aktuell durchnummeriert)
			
			segments = int(skeletonarea.max()) #new number of segments after trimming
			junctions = len(linejunctions) #number of junctions in the picture
			
			if silent == 0:
				print("segments:",segments,"junctions:",junctions)
			
			#------- eliminate segments, test if many, few or no junctions
			wormcore = np.zeros((segmentedskeleton.shape)) #leeres array in der größe des skeletonbildes anlegen
			
			if (segments <= 1 or junctions == 0): #only one segment, so no trouble here
				if silent == 0:
					print("---no fragmentation")
				wormcore = segmentedskeleton.copy()
			
			#------- few junctions - just keep the longest ones
			elif ((2 <= segments <= 6) or (junctions <= 3)): 
				if silent == 0:
					print("---low fragmentation")
				segmentlist = np.zeros((segments))
				for i in range (0,segments):
					segmentlist[i] = ndimage.sum(segmentedskeleton == (i + 1))
				for i in range (0,segments):
					if (segmentlist[i] <= (sum(segmentlist) * 0.10)):	#keep if mass >10%
						segmentedskeleton[segmentedskeleton == i + 1] = 0
			
				segmentedskeleton[segmentedskeleton > 1] = 1
				wormcore = segmentedskeleton.copy()
			
			#------ many junctions - discard all segments that don't have junctions at each end but keep those above 15% total mass
			elif (junctions > 3):
				if silent == 0:
					print("---highly fragmented")
				for i in range(1,segments + 1):
					seg_array,seg_coord = HMlineends((segmentedskeleton == i))
					d1_smallest = 4 #maximale distanz bei disk(2) = 3.16
					d2_smallest = 4
					for j in range(0,junctions):
						a = seg_coord[0,0] - linejunctions[j,0]
						b = seg_coord[0,1] - linejunctions[j,1]
						d1 = hypot(a,b)
						a = seg_coord[1,0] - linejunctions[j,0]
						b = seg_coord[1,1] - linejunctions[j,1]
						d2 = hypot(a,b)
						
						if (d1 < d1_smallest):
							d1_smallest = d1	
						if (d2 < d2_smallest):
							d2_smallest = d2
					if (d1_smallest < 4 and d2_smallest < 4):
						wormcore += (segmentedskeleton == i)
			
				segmentlist = np.zeros((segments))
				for i in range (0,segments):
					segmentlist[i] = ndimage.sum(segmentedskeleton == (i + 1))
				for i in range (0,segments):
					if (segmentlist[i] <= (sum(segmentlist) * 0.15)): #15% mass
						segmentedskeleton[segmentedskeleton == i + 1] = 0
				wormcore += segmentedskeleton
				wormcore[wormcore > 1] = 1
			#------
			
			if (showplots == 1):
				#show extracted segments
				subplot(zeilen,spalten,8)
				imshow(wormcore,interpolation='none',cmap='jet')
				plt.axis('off')
				title("remaining segments")
				
				#linejunctions and lineends bunt anmalen
				segmentedlineendsarray, segmentedlineendscoord = HMlineends(segmentedskeleton)
				linejuncpic = skeletonwurm  + linejunctionwurm * 3 + segmentedlineendsarray * 5
				subplot(zeilen,spalten,7)
				imshow(linejuncpic,interpolation='none',cmap='jet')
				plt.axis('off')
				title("junctions and lineends")
			
			segmentedlineendsarray, segmentedlineendscoord = HMlineends(segmentedskeleton)
			
			if (timings == 1):
				end = time.time()
				print("segmentation",end - start)
				start = end
			
			############################################################
			############################################################
			#------ NOSEFINDER -----------------------------------------
			
			nose = np.array([0,0])
			tail = np.array([0,0])
			realnose = np.array([0,0])
			
			nose[0],nose[1],tail[0],tail[1] = nosefinder(wurm,wormcore,offsetx,offsety,closedwurm,newblock,lastnose)
			
			lastnose = nose.copy()
			lasttail = tail.copy()
			
			realnose[0] = nose[0]
			realnose[1] = nose[1]
			
			############################################################
			############################################################
			#If nose jumped repeat frame, also check entropy
			#if newblock == 0:
			#	a = realnose[0] - lastrealnose[0]
			#	b = realnose[1] - lastrealnose[1]
			#	c = hypot(a,b)
			#	if c > middle_distance: #100 ~ half the head
			#		print("Nose jumped! (Probably) in frame",cframe)
			#		cframe -= 1
			#		force_newblock = 1
				
			lastrealnose = realnose
			
			#------ lineends bunt anmalen
			realnosetail = wurm.copy()
			
			#old black and white
			#time: 0.1436 --> 500
			#realnosetail[tail[0] - 3:tail[0] + 3,tail[1] - 3:tail[1] + 3] = white_reference/2
			#realnosetail[realnose[0],realnose[1]] = white_reference
			#realnosetail[(realnose[0]) - 2:(realnose[0]) + 2,(realnose[1]) - 2:(realnose[1]) + 2] = white_reference
			
			#--- RGB transformation
			#time: 6.4830 --> 500
			color_realnosetail = np.zeros((realnosetail.shape[0],realnosetail.shape[1],3))#, dtype=np.uint8)
			
			color_realnosetail[:,:,0] = realnosetail
			color_realnosetail[:,:,1] = realnosetail
			color_realnosetail[:,:,2] = realnosetail
			
			#needs black refernce for appropriate color stretching
			black_reference = realnosetail.min()
			
			#adding color points - delete everything except specific layer RGB
			color_realnosetail[tail[0] - 3:tail[0] + 3,tail[1] - 3:tail[1] + 3] = [black_reference,white_reference/2,white_reference/2]
			
			color_realnosetail[realnose[0],realnose[1]] = [white_reference,black_reference,black_reference]
			color_realnosetail[(realnose[0]) - 2:(realnose[0]) + 2,(realnose[1]) - 2:(realnose[1]) + 2] = [white_reference,black_reference,black_reference]
			
			color_realnosetail[com[0] - 3:com[0] + 3,com[1] - 3:com[1] + 3] = [black_reference,white_reference/2,black_reference]
			
			if (showplots == 1):
				subplot(zeilen,spalten,1)
				imshow(realnosetail,interpolation='none',cmap='binary')
				plt.axis('off')
				title(wurmsource)
				
				subplot(zeilen,spalten,3)
				#imshow(nosesegment+closedwurm,interpolation='none',cmap='binary')
				imshow(closedwurm,interpolation='none',cmap='binary')
				plt.axis('off')
				title(wurmsource)
				
			#------
			if (timings == 1):
				end = time.time()
				print("match function",end - start)
				start = end
			
			####################################################################
			#------ write image to disc ----------------------------------------
			
			if (force_newblock == 1):
				temp_cframe = cframe + 1 #do this because of string/float conversion in next line
				wurmout = directory + '/Nose_pictures/out_NT_%05d.tif' % temp_cframe
			else:
				wurmout = directory + '/Nose_pictures/out_NT_%05d.tif' % cframe
			
			if (force_newblock == 0):
				outfile.write(str(cframe) + "," + str(realnose[0]) + "," + str(realnose[1]) + "," + str(tail[0]) + "," + str(tail[1]) + "," + str(com[0]) + "," + str(com[1]) + "\n")
			
			scipy.misc.imsave(wurmout,color_realnosetail)
			
			#----- layout plots
			if (showplots == 1):
				subplots_adjust(left=0.015, bottom=0.015, right=0.985, top=0.99, wspace=0.07, hspace=0.025)
				show()
				close('all')
			####################################################################
			#----- finish it up
			cframe += 1
		
		outfile.close()
		stop = time.time()
		#print("blocktime",start_frame,"-",stop_frame,"%.2f" % (((stop_frame-start_frame)/(stop-start))*60),"fpm")

	except Exception as e:
		print(e)
		print(">>> error in frame:",cframe)

def wurmstart(startframe,endframe,pcrop,directory,blocksize,framecount,targetsize,bestedge,startnoseposition,Globalstart,qposition,c_qposition,px_size,frame_time):
	global total_chunks
	#---set blocksize
	global block
	global blockchange
	block = list(xfrange(1,framecount + 1,blocksize)) #list where to create a new mulitprocssing block at least
	blockchange = list(xfrange(1,framecount + 1,blocksize)) #list where burst mode ends
	
	#print("New Blocks at frames:",block)
	
	#---create data directories
	try:
		os.stat(directory + "/Tracking_data")
	except:
		os.mkdir(directory + "/Tracking_data")
	try:
		os.stat(directory + "/Nose_pictures")
	except:
		os.mkdir(directory + "/Nose_pictures")
	
	#---timing stuff
	allstart = time.time()
	duration = (endframe - startframe)/ntc.nt_currentfpm
	qprogress = c_qposition - 1
	qsize = 54
	qfactor = qsize/(qposition -1)
	
	print("-----------------------------------------------------------------------")
	print("Total Progress:","[","#"*math.ceil(qprogress*qfactor),"-"*(qsize-math.ceil(qprogress*qfactor)),"]",sep="")
	print("Estimated Block Time:","%.1f" % duration,"min.")
	print(directory)
	print("-----------------------------------------------------------------------")
	
	#------ Main program ---------------------------
	global pool
	pool = Pool(processes=ntc.nt_parallel_processes)
	
	total_chunks = multicorewurm(startframe,endframe,blockchange,pcrop,directory,targetsize,bestedge,startnoseposition,Globalstart)
	
	pool.close()
	pool.join()
	#-----------------------------------------------
	
	#---timing stuff
	allend = time.time()
	
	a = total_chunks + 2 #move cursor down number of chunks + empty line
	
	print("\n"*a,sep="")
	print("---------------------------------------------------------------")
	print("block time elapsed:","%.2f" % ((allend - allstart)/60),"min.")
	fpm = ((endframe - startframe)/(allend - allstart))*60
	print("pictures per minute:","%.1f" % fpm)
	
	#----- merge all output files into single data file and draw informative graphs
	wormname = directory.split('/')[-2].replace(" ","_")
	
	filenames = sorted(glob.glob(directory + '/Tracking_data/*.txt'))
	with open(directory + '/Tracking_data/' + wormname + '_Nose.dat', 'w') as mergefile:
		mergefile.write("#" + str("Nose Speed " + directory.split('/')[-2]) + "\n")
		mergefile.write("#Frame No.,Nose X, Nose Y,Tail X, Tail Y,Centroid X,Centroid Y\n")
		for fname in filenames:
			with open(fname) as infile:
				for line in infile:
					mergefile.write(line)
					
	#---delete txt files
	filelist = glob.glob(directory + "/Tracking_data/*.txt")
	for f in filelist:
		os.remove(f)
	
	#---delte working pictures
	filelist = glob.glob(directory + "/*.tif")
	for f in filelist:
		os.remove(f)
	
	#-------------------------------------------------------------------
	#---save figures, nose + centroid
	#create file containning movement speed of nose and centroid
	datafile = np.loadtxt(directory + '/Tracking_data/' + wormname + '_Nose.dat',delimiter=",") #load datafile into numpy array
	movedata = np.zeros(((len(datafile[:,0]),3))) #create empty datatable for nose and centroid moving data, also for plots [frame,nosespeed,centroidspeed]
	movedata[:,0] = datafile[:,0] #copy frame numbers
	
	#calculate nose and centroid moving distance
	for i in range(0,len(datafile[:,0])-1):
		a = datafile[i,1] - datafile[i+1,1]
		b = datafile[i,2] - datafile[i+1,2]
		movedata[i,1] = hypot(a,b) * px_size / frame_time
		
		a = datafile[i,5] - datafile[i+1,5]
		b = datafile[i,6] - datafile[i+1,6]
		movedata[i,2] = hypot(a,b) * px_size / frame_time		
		
	#add nose and centroid speed to the table
	alldata = np.zeros((datafile.shape[0],datafile.shape[1]+2))
	for i in range(0,len(datafile[:,0])):
		for j in range(0,datafile.shape[1]):
			alldata[i,j] = datafile[i,j]
		alldata[i,datafile.shape[1]] = movedata[i,1]  #adds nose speed 
		alldata[i,datafile.shape[1]+1] = movedata[i,2] #adds centroid speed
	
	#save table with calculated speed values
	np.savetxt(directory + '/Tracking_data/' + wormname + '_Nose.dat',alldata,delimiter=',',header="Nose Speed " + wormname + "\nFrame No.,Nose X, Nose Y,Tail X, Tail Y,Centroid X,Centroid Y,Nose Speed,Centroid Speed")
	
	#-------------------------------------------------------------------
	#Nose speed graphic
	#try to calculate mean values
	try:
		block1mean,block2mean,block3mean = blockmean(ntc.block1,ntc.block2,ntc.block3,movedata,1)
		meanValues = "Block 1: " + str(block1mean) + " (" + str(ntc.block1) + "frames)" + "\n" + "Block 2: " + str(block2mean) + " (" + str(ntc.block2) + "frames)" + "\n" + "Block 3: " + str(block3mean) + " (" + str(ntc.block3) + " frames)"
	except:
		meanValues = "not enough data\nfor mean values"
	
	#draw and save graph for nose speed
	Graphlabel = "Nose Speed " + directory.split('/')[-2]
	plt.title(Graphlabel)
	plt.xlabel('Frame #')
	plt.xticks(np.arange(min(movedata[:,0]),max(movedata[:,0]),40)-1)
	plt.ylabel('Speed: $\mu$m/s')
	plt.ylim((0,100))
	plt.annotate(meanValues,xy=(0.05,0.9), xycoords='axes fraction', xytext=(0,0), textcoords='offset points',ha='left', va='center')
	plt.plot(movedata[:,0],movedata[:,1], 'b.', label="speed")
	savefig(directory + '/Tracking_data/' + wormname + '_NT_nose_speed.pdf', bbox_inches='tight')
	plt.close()
	
	#Centroid speed graphic
	#try to calculate mean values
	try:
		block1mean,block2mean,block3mean = blockmean(ntc.block1,ntc.block2,ntc.block3,movedata,2)
		meanValues = "Block 1: " + str(block1mean) + "\n" + "Block 2: " + str(block2mean) + "\n" + "Block 3: " + str(block3mean)
	except:
		meanValues = "not enough data\nfor mean values"
	
	#draw and save graph for centroid speed
	Graphlabel = "Centroid Speed " + directory.split('/')[-2]
	plt.title(Graphlabel)
	plt.xlabel('Frame #')
	plt.xticks(np.arange(min(movedata[:,0]),max(movedata[:,0]),40)-1)
	plt.ylabel('Speed: $\mu$m/s')
	plt.ylim((ntc.y_minimum,ntc.y_maximum))
	plt.annotate(meanValues,xy=(0.05,0.9), xycoords='axes fraction', xytext=(0,0), textcoords='offset points',ha='left', va='center')
	plt.plot(movedata[:,0],movedata[:,2], 'g.', label="speed")
	savefig(directory + '/Tracking_data/' + wormname + '_NT_centroid_speed.pdf', bbox_inches='tight')
	plt.close()
	
	
	#-------------------------------------------------------------------
	#save datafile and graphs to shared folder
	if ntc.shared_folder == 1:
		try:
			os.stat(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		except:
			os.mkdir(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		
		shutil.copy2(directory + '/Tracking_data/' + wormname + '_Nose.dat',os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data/")
	if ntc.shared_plots == 1:
		try:
			os.stat(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		except:
			os.mkdir(os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data")
		
		shutil.copy2(directory + '/Tracking_data/' + wormname + '_NT_nose_speed.pdf',os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data/")
		shutil.copy2(directory + '/Tracking_data/' + wormname + '_NT_centroid_speed.pdf',os.path.dirname(os.path.dirname(directory)) + "/" + directory.split('/')[-3] + "_Collected_data/")
			






if __name__ == '__main__':
	
	#(startframe,endframe,[c,r,o,p],'directory',blocksize,max # of frames,targetsize)
	bestedge = scipy.misc.imread('/home/birk/Videos/temp/bestedge.tif')
	
	givemeframes = 321
	
	startnoseposition = np.array([10, 313, 246, 270, 332, 132, 90, 199])
	
	wurmstart(givemeframes,325,[41,109,66,101],'/home/birk/Videos/150621_HBR592_field2_20150622_21221 PM//working/',120,480,5600,bestedge,startnoseposition)
	
	
#?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?
#?!?!?! NOTES ?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!
#?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?

	#~ #print("Memory:",(resource.getrusage(resource.RUSAGE_SELF).ru_maxrss)/1024,"MB")
	#~ #objgraph.show_growth(limit=3)

	#scipy.ndimage.morphology.binary_fill_holes
	#einen blick wert

#?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?!?

########################################################################
####### HERE BE SNIPPETS ###############################################
#defaults:
#left  = 0.125  # the left side of the subplots of the figure
#right = 0.9    # the right side of the subplots of the figure
#bottom = 0.1   # the bottom of the subplots of the figure
#top = 0.9      # the top of the subplots of the figure
#wspace = 0.2   # the amount of width reserved for blank space between subplots
#hspace = 0.5   # the amount of height reserved for white space between subplots	
